@@ -1,4 +1,4 @@
-/* Copyright (C) 2006 Massachusetts Institute of Technology
+/* Copyright (C) 2005-2008 Massachusetts Institute of Technology
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -31,6 +31,10 @@
 #include "config.h"
 
 #ifdef HAVE_HDF5
+
+/* don't use new HDF5 1.8 API (which isn't even fully documented yet, grrr) */
+#  define H5_USE_16_API 1
+
 #  include <hdf5.h>
 
 /* HDF5 changed this datatype in their interfaces starting in version 1.6.4 */
@@ -66,13 +70,12 @@ typedef int hid_t;
 #  endif
 #endif
 
-#ifdef HAVE_H5PSET_FAPL_MPIO
+#if defined(HAVE_H5PSET_FAPL_MPIO) || !defined(HAVE_MPI)
 #  define IF_EXCLUSIVE(yes,no) no
 #else
 #  define IF_EXCLUSIVE(yes,no) yes
-#endif
-
 static int h5io_critical_section_tag = 0;
+#endif
 
 /*****************************************************************************/
 /* Normally, HDF5 prints out all sorts of error messages, e.g. if a dataset
@@ -113,12 +116,14 @@ void *h5file::get_id() {
 
 #ifdef HAVE_HDF5
     hid_t access_props = H5Pcreate (H5P_FILE_ACCESS);
-#  if defined(HAVE_MPI) && defined(HAVE_H5PSET_FAPL_MPIO)
+#  ifdef HAVE_MPI
+#    ifdef HAVE_H5PSET_FAPL_MPIO
     if (parallel)
       H5Pset_fapl_mpio(access_props, MPI_COMM_WORLD, MPI_INFO_NULL);
-#  else
+#    else
     if (parallel)
       begin_critical_section(h5io_critical_section_tag);
+#    endif
 #  endif
 
     if (mode != WRITE || IF_EXCLUSIVE(parallel && !am_master(), 0))
@@ -137,28 +142,30 @@ void *h5file::get_id() {
 // hackery: in some circumstances, for the exclusive-access mode
 // we must close the id (i.e. the file) in order to prevent deadlock.
 void h5file::prevent_deadlock() {
-  IF_EXCLUSIVE(if (parallel) close_id(), 0);
+  IF_EXCLUSIVE(if (parallel) close_id(), (void) 0);
 }
 
 void h5file::close_id() {
   unset_cur();
-#ifdef HAVE_HDF5
   if (HID(id) >= 0)
+    if (mode == WRITE) mode = READWRITE; // don't re-create on re-open
+#ifdef HAVE_HDF5
+  if (HID(id) >= 0) {
     H5Fclose(HID(id));
-  HID(id) = -1;
+    IF_EXCLUSIVE(if (parallel)
+		 end_critical_section(h5io_critical_section_tag++),
+		 (void) 0);
+  }
 #endif
-  IF_EXCLUSIVE(if (parallel)
-	       end_critical_section(h5io_critical_section_tag++),
-	       (void) 0);
-  if (mode == WRITE) mode = READWRITE; // don't re-create on re-open
+  HID(id) = -1;
 }
 
 /* note: if parallel is true, then *all* processes must call this,
    and all processes will use I/O. */
 h5file::h5file(const char *filename_, access_mode m, bool parallel_) {
   cur_dataname = NULL;
-  id = (void*) (new hid_t);
-  cur_id = (void*) (new hid_t);
+  id = (void*) malloc(sizeof(hid_t));
+  cur_id = (void*) malloc(sizeof(hid_t));
   HID(id) = -1;
   HID(cur_id) = -1;
   extending = 0;
@@ -173,11 +180,13 @@ h5file::~h5file() {
   if (cur_dataname) free(cur_dataname); // allocated with realloc
   for (h5file::extending_s *cur = extending; cur; ) {
     h5file::extending_s *next = cur->next;
-    delete cur->dataname;
+    delete[] cur->dataname;
     delete cur;
     cur = next;
   }
   delete[] filename;
+  free(cur_id);
+  free(id);
 }
 
 bool h5file::ok() {
@@ -189,13 +198,13 @@ void h5file::remove() {
   if (mode == READWRITE) mode = WRITE; // now need to re-create file
   for (h5file::extending_s *cur = extending; cur; ) {
     h5file::extending_s *next = cur->next;
-    delete cur->dataname;
+    delete[] cur->dataname;
     delete cur;
     cur = next;
   }
   extending = 0;
   
-  IF_EXCLUSIVE(if (parallel) all_wait(), 0);
+  IF_EXCLUSIVE(if (parallel) all_wait(), (void) 0);
   if (am_master() && std::remove(filename))
     abort("error removing file %s", filename);
 }
@@ -404,7 +413,7 @@ void h5file::remove_data(const char *dataname)
       prev->next = cur->next;
     else
       extending = cur->next;
-    delete cur->dataname;
+    delete[] cur->dataname;
     delete cur;
   }
 
@@ -472,6 +481,7 @@ void h5file::create_data(const char *dataname, int rank, const int *dims,
     hid_t type_id = single_precision ? H5T_NATIVE_FLOAT : H5T_NATIVE_DOUBLE;
     
     data_id = H5Dcreate(file_id, dataname, type_id, space_id, prop_id);
+    if (data_id < 0) abort("Error creating dataset");
     
     H5Pclose(prop_id);
   }
@@ -696,7 +706,7 @@ void h5file::write(const char *dataname, int rank, const int *dims,
 void h5file::write(const char *dataname, const char *data)
 {
 #ifdef HAVE_HDF5
-  if (parallel || am_master()) {
+  if (am_master()) { // only master process writes string data
     hid_t file_id = HID(get_id()), type_id, data_id, space_id;
     
     CHECK(file_id >= 0, "error opening HDF5 output file");     
