@@ -1,4 +1,4 @@
-/* Copyright (C) 2006 Massachusetts Institute of Technology
+/* Copyright (C) 2005-2008 Massachusetts Institute of Technology
 %
 %  This program is free software; you can redistribute it and/or modify
 %  it under the terms of the GNU General Public License as published by
@@ -29,11 +29,7 @@ extern bool quiet; // if true, suppress all non-error messages from Meep
 
 const double pi = 3.141592653589793238462643383276;
 
-#ifdef INFINITY
-const double infinity = INFINITY;
-#else
-const double infinity = 1e20; // this should be big enough for us
-#endif
+const double infinity = HUGE_VAL;
 
 #ifdef NAN
 const double nan = NAN;
@@ -123,11 +119,13 @@ private:
   void close_id();
 };
 
+typedef double (*pml_profile_func)(double u, void *func_data);
+
 #define DEFAULT_SUBPIXEL_TOL 1e-4
 #define DEFAULT_SUBPIXEL_MAXEVAL 100000
 
 /* This class is used to compute position-dependent material properties
-   like the dielectric function, polarizability sigma, 
+   like the dielectric function, permeability (mu), polarizability sigma, 
    nonlinearities, et cetera.  Simple cases of stateless functions are
    handled by canned subclasses below, but more complicated cases
    can be handled by creating a user-defined subclass of material_function.
@@ -135,7 +133,7 @@ private:
    it is likely that complicated implementations will share state between
    properties. */
 class material_function {
-  material_function(const material_function &ef) {} // prevent copying
+  material_function(const material_function &ef) {(void)ef;} // prevent copying
 public:
   material_function() : omega(nan), gamma(nan), deps(nan), energy_sat(nan) {}
   
@@ -143,12 +141,21 @@ public:
   
   /* Specify a restricted volume: all subsequent eps/sigma/etc
      calls will be for points inside gv, until the next set_volume. */
-  virtual void set_volume(const geometric_volume &gv) {}
+  virtual void set_volume(const geometric_volume &gv) {(void)gv;}
   virtual void unset_volume(void) {} // unrestrict the volume
   
   /* scalar dielectric function */
-  virtual double eps(const vec &r) { return 1.0; }
+  virtual double eps(const vec &r) { (void)r; return 1.0; }
+
+  /* scalar permeability function */
+  virtual bool has_mu() { return false; } /* true if mu != 1 */
+  virtual double mu(const vec &r) { (void)r; return 1.0; }
   
+  /* scalar conductivity function */
+  virtual bool has_conductivity(component c) { (void)c; return false; }
+  virtual double conductivity(component c, const vec &r) { 
+    (void) c; (void)r; return 0.0; }  
+
   /* Return interface normal and/or average dielectric eps and 1/eps
      in a given volume gv.   These are virtual so that e.g. libctl
      can override them with more efficient geometry-based routines. */
@@ -158,9 +165,8 @@ public:
 		       double tol=DEFAULT_SUBPIXEL_TOL, 
 		       int maxeval=DEFAULT_SUBPIXEL_MAXEVAL);
   
-
   /* polarizability sigma function */
-  virtual double sigma(const vec &r) { return 0.0; }
+  virtual double sigma(const vec &r) { (void)r; return 0.0; }
   /* specify polarizability used for subsequent calls to sigma(r) */
   virtual void set_polarizability(double omega_, double gamma_, 
 				  double deps_, double energy_sat_) {
@@ -169,9 +175,9 @@ public:
   
   // Nonlinear susceptibilities
   virtual bool has_chi3() { return false; }
-  virtual double chi3(const vec &r) { return 0.0; }
+  virtual double chi3(const vec &r) { (void)r; return 0.0; }
   virtual bool has_chi2() { return false; }
-  virtual double chi2(const vec &r) { return 0.0; }
+  virtual double chi2(const vec &r) { (void)r; return 0.0; }
   
   // TODO: dielectric tensor, ...
 
@@ -189,6 +195,9 @@ public:
   virtual ~simple_material_function() {}
   
   virtual double eps(const vec &r) { return f(r); }
+  virtual double mu(const vec &r) { return f(r); }
+  virtual double conductivity(component c, const vec &r) { 
+    (void)c; return f(r); }
   virtual double sigma(const vec &r) { return f(r); }
   virtual double chi3(const vec &r) { return f(r); }
   virtual double chi2(const vec &r) { return f(r); }
@@ -201,8 +210,12 @@ class structure_chunk {
   double a, Courant, dt; // res. a, Courant num., and timestep dt=Courant/a
   double *eps, *chi3[NUM_FIELD_COMPONENTS], *chi2[NUM_FIELD_COMPONENTS];
   double *inveps[NUM_FIELD_COMPONENTS][5];
-  double *C[5][NUM_FIELD_COMPONENTS];
-  double *Cdecay[5][NUM_FIELD_COMPONENTS][5];
+  double *invmu[NUM_FIELD_COMPONENTS][5];
+  double *conductivity[NUM_FIELD_COMPONENTS][5];
+  double *condinv[NUM_FIELD_COMPONENTS][5]; // cache of 1/(1+conduct*dt/2)
+  bool condinv_stale; // true if condinv needs to be recomputed
+  double *sig[5], *siginv[5]; // conductivity array for uPML
+  int sigsize[5]; // conductivity array size
   volume v;  // integer volume that could be bigger than non-overlapping gv below
   geometric_volume gv;
   polarizability *pb;
@@ -216,11 +229,14 @@ class structure_chunk {
   void set_epsilon(material_function &eps,
                    bool use_anisotropic_averaging,
 		   double tol, int maxeval);
+  void set_mu(material_function &eps);
+  void set_conductivity(component c, material_function &eps);
+  void update_condinv();
   void set_chi3(material_function &eps);
   void set_chi2(material_function &eps);
-  void use_pml(direction, double dx, double boundary_loc, double strength);
-  void update_pml_arrays();
-  void update_Cdecay();
+  void use_pml(direction, double dx, double boundary_loc, double Rasymptotic, 
+	       pml_profile_func pml_profile, void *pml_profile_data,
+	       double pml_profile_integral);
 
   void add_polarizability(double sigma(const vec &), double omega, double gamma,
                           double delta_epsilon = 1.0, double energy_saturation = 0.0);
@@ -243,23 +259,27 @@ class structure_chunk {
   int the_is_mine;
 };
 
+double pml_quadratic_profile(double, void*);
+
 // linked list of descriptors for boundary regions (currently just for PML)
 class boundary_region {
 public:
   typedef enum { NOTHING_SPECIAL, PML } boundary_region_kind;
   
   boundary_region() :
-    kind(NOTHING_SPECIAL), thickness(0.0), strength(1.0), d(NO_DIRECTION), side(Low), next(0) {}
-  boundary_region(boundary_region_kind kind, double thickness, double strength, direction d, boundary_side side, boundary_region *next = 0) : kind(kind), thickness(thickness), strength(strength), d(d), side(side), next(next) {}
+    kind(NOTHING_SPECIAL), thickness(0.0), Rasymptotic(1e-16), pml_profile(NULL), pml_profile_data(NULL), pml_profile_integral(1.0), d(NO_DIRECTION), side(Low), next(0) {}
+  boundary_region(boundary_region_kind kind, double thickness, double Rasymptotic, pml_profile_func pml_profile, void* pml_profile_data, double pml_profile_integral, direction d, boundary_side side, boundary_region *next = 0) : kind(kind), thickness(thickness), Rasymptotic(Rasymptotic), pml_profile(pml_profile), pml_profile_data(pml_profile_data), pml_profile_integral(pml_profile_integral), d(d), side(side), next(next) {}
 
-  boundary_region(const boundary_region &r) : kind(r.kind), thickness(r.thickness), strength(r.strength), d(r.d), side(r.side) { 
+  boundary_region(const boundary_region &r) : kind(r.kind), thickness(r.thickness), Rasymptotic(r.Rasymptotic), pml_profile(r.pml_profile), pml_profile_data(r.pml_profile_data), pml_profile_integral(r.pml_profile_integral), d(r.d), side(r.side) { 
     next = r.next ? new boundary_region(*r.next) : 0;
   }
 
   ~boundary_region() { if (next) delete next; }
   
   void operator=(const boundary_region &r) {
-    kind = r.kind; thickness = r.thickness; strength = r.strength;
+    kind = r.kind; thickness = r.thickness; Rasymptotic = r.Rasymptotic;
+    pml_profile = r.pml_profile; pml_profile_data = r.pml_profile_data;
+    pml_profile_integral = r.pml_profile_integral;
     d = r.d; side = r.side;
     if (next) delete next;
     next = r.next ? new boundary_region(*r.next) : 0;
@@ -274,7 +294,7 @@ public:
   boundary_region operator*(double strength_mult) const {
     boundary_region r(*this), *cur = &r;
     while (cur) {
-      cur->strength *= strength_mult;
+      cur->Rasymptotic = pow(cur->Rasymptotic, strength_mult);
       cur = cur->next;
     }
     return r;
@@ -285,7 +305,10 @@ public:
 
 private:
   boundary_region_kind kind;
-  double thickness, strength;
+  double thickness, Rasymptotic;
+  pml_profile_func pml_profile;
+  void *pml_profile_data;
+  double pml_profile_integral;
   direction d;
   boundary_side side;
   boundary_region *next;
@@ -340,6 +363,10 @@ class structure {
                    bool use_anisotropic_averaging=true,
 		   double tol=DEFAULT_SUBPIXEL_TOL,
 		   int maxeval=DEFAULT_SUBPIXEL_MAXEVAL);
+  void set_mu(material_function &eps);
+  void set_mu(double mu(const vec &));
+  void set_conductivity(component c, material_function &conductivity);
+  void set_conductivity(component C, double conductivity(const vec &));
   void set_chi3(material_function &eps);
   void set_chi3(double eps(const vec &));
   void set_chi2(material_function &eps);
@@ -368,7 +395,7 @@ class structure {
   friend class boundary_region;
 
  private:
-  void use_pml(direction d, boundary_side b, double dx, double strength);
+  void use_pml(direction d, boundary_side b, double dx);
   void add_to_effort_volumes(const volume &new_effort_volume, 
 			     double extra_effort);
   void choose_chunkdivision(const volume &v, int num_chunks,
@@ -388,9 +415,17 @@ class flux_vol;
 // current = d(dipole)/dt (or rather, the finite-difference equivalent).
 class src_time {
  public:
-  src_time() { current_time = nan; current_current = 0.0; next = NULL; }
+  // the following variable specifies whether the current
+  // source is specified as a current or as an integrated
+  // current (a dipole moment), if possible.  In the original Meep,
+  // by default electric sources are integrated and magnetic
+  // sources are not, but this may change.
+  bool is_integrated;
+
+  src_time() { is_integrated = true; current_time = nan; current_current = 0.0; next = NULL; }
   virtual ~src_time() { delete next; }
   src_time(const src_time &t) { 
+       is_integrated = t.is_integrated;
        current_time = t.current_time;
        current_current = t.current_current;
        current_dipole = t.current_dipole;
@@ -407,7 +442,10 @@ class src_time {
     }
   }
 
-  complex<double> current(double time, double dt) const { 
+  // subclasses *can* override this method in order to specify the
+  // current directly rather than as the derivative of dipole.
+  // in that case you would probably ignore the dt argument.
+  virtual complex<double> current(double time, double dt) const { 
     return ((dipole(time + dt) - dipole(time)) / dt);
   }
 
@@ -476,7 +514,12 @@ class custom_src_time : public src_time {
     : func(func), data(data), start_time(st), end_time(et) {}
   virtual ~custom_src_time() {}
   
-  virtual complex<double> dipole(double time) const { return func(time,data); }
+  virtual complex<double> current(double time, double dt) const { 
+    if (is_integrated) return src_time::current(time,dt);
+    else return dipole(time);
+  }
+  virtual complex<double> dipole(double time) const { 
+    if (time >= start_time) return func(time,data); else return 0.0; }
   virtual double last_time() const { return end_time; };
   virtual src_time *clone() const { return new custom_src_time(*this); }
   virtual bool is_equal(const src_time &t) const;
@@ -592,7 +635,7 @@ public:
 
   void scale_dfts(complex<double> scale);
 
-  void remove() { delete E; delete H; E = H = 0; }
+  void remove();
 
   double freq_min, dfreq;
   int Nfreq;
@@ -605,15 +648,18 @@ enum connect_phase { CONNECT_PHASE = 0, CONNECT_NEGATE=1, CONNECT_COPY=2 };
 
 class fields_chunk {
  public:
-  double *f[NUM_FIELD_COMPONENTS][2];
-  double *f_backup[NUM_FIELD_COMPONENTS][2];
-  double *f_p_pml[NUM_FIELD_COMPONENTS][2];
-  double *f_m_pml[NUM_FIELD_COMPONENTS][2];
-  double *f_backup_p_pml[NUM_FIELD_COMPONENTS][2];
-  double *f_backup_m_pml[NUM_FIELD_COMPONENTS][2];
+  double *f[NUM_FIELD_COMPONENTS][2]; // fields at current time
+  double *f_prev[NUM_FIELD_COMPONENTS][2]; // field at prev step (used in PML)
 
-  bool have_d_minus_p;
-  double *d_minus_p[NUM_FIELD_COMPONENTS][2];
+  /* sometimes, to synchronize the E and H fields, e.g. for computing
+     flux at a given time, we need to timestep H by 1/2; in this case
+     we save backup copies of (some of) the fields to resume timestepping */
+  double *f_backup[NUM_FIELD_COMPONENTS][2];
+
+  // used to store D-P and B-P, e.g. when P implements dispersive media
+  double *f_minus_p[NUM_FIELD_COMPONENTS][2];
+
+  double *f_rderiv_int; // cache of helper field for 1/r d(rf)/dr derivative
 
   dft_chunk *dft_chunks;
 
@@ -628,14 +674,14 @@ class fields_chunk {
   volume v;
   geometric_volume gv;
   double m, rshift;
-  int is_real;
+  int is_real, store_pol_energy;
   bandsdata *bands;
-  src_vol *e_sources, *h_sources;
+  src_vol *d_sources, *b_sources;
   const structure_chunk *new_s;
   structure_chunk *s;
   const char *outdir;
 
-  fields_chunk(structure_chunk *, const char *outdir, double m);
+  fields_chunk(structure_chunk *, const char *outdir, double m, bool store_pol_energy);
   fields_chunk(const fields_chunk &);
   ~fields_chunk();
 
@@ -645,7 +691,7 @@ class fields_chunk {
   void use_real_fields();
   bool have_component(component c, bool is_complex = false) {
     switch (c) {
-    case Dielectric:
+    case Dielectric: case Permeability:
       return !is_complex;
     default:
       return (f[c][0] && f[c][is_complex]);
@@ -670,9 +716,10 @@ class fields_chunk {
   double get_eps(const ivec &iloc) const;
   complex<double> analytic_epsilon(double freq, const vec &) const;
   
-  void backup_h();
-  void restore_h();
-
+  void backup_component(component c);
+  void average_with_backup(component c);
+  void restore_component(component c);
+  
   void set_output_directory(const char *name);
   void verbose(int v=1) { verbosity = v; }
 
@@ -693,6 +740,9 @@ class fields_chunk {
   void remove_polarizabilities();
   void zero_fields();
 
+  bool update_h_from_b();
+  void update_e_from_d();
+
  private: 
   int verbosity; // Turn on verbosity for debugging purposes...
   // fields.cpp
@@ -708,11 +758,9 @@ class fields_chunk {
   // step.cpp
   void phase_in_material(const structure_chunk *s);
   void phase_material(int phasein_time);
-  void step_h();
-  void step_h_source(src_vol *);
-  void step_d();
-  void step_d_source(src_vol *);
-  void update_e_from_d();
+  void step_b_source(src_vol *, int including_integrated);
+  void step_db(field_type ft);
+  void step_d_source(src_vol *, int including_integrated);
   void update_from_e();
   void calc_sources(double time);
 
@@ -748,7 +796,7 @@ typedef double (*field_rfunction)(const complex<double> *fields,
 				   void *integrand_data_);
 
 field_rfunction derived_component_func(derived_component c, const volume &v,
-				       int &nfields, component cs[6]);
+				       int &nfields, component cs[12]);
 
 class fields {
  public:
@@ -780,7 +828,7 @@ class fields {
   bandsdata *bands;
   char *outdir;
   // fields.cpp methods:
-  fields(structure *, double m=0);
+  fields(structure *, double m=0, bool store_pol_energy=0);
   fields(const fields &);
   ~fields();
   bool equal_layout(const fields &f) const;
@@ -949,6 +997,8 @@ class fields {
   double get_field(derived_component c, const vec &loc) const;
 
   // energy_and_flux.cpp
+  void synchronize_magnetic_fields();
+  void restore_magnetic_fields();
   double energy_in_box(const geometric_volume &);
   double electric_energy_in_box(const geometric_volume &);
   double magnetic_energy_in_box(const geometric_volume &);
@@ -982,13 +1032,14 @@ class fields {
   // material.cpp
   double max_eps() const;
   // step.cpp
-  void force_consistency(field_type ft);
+  void step_boundaries(field_type);
 
   bool nosize_direction(direction d) const;
   direction normal_direction(const geometric_volume &where) const;
 
  private: 
   int verbosity; // Turn on verbosity for debugging purposes...
+  int synchronized_magnetic_fields; // count number of nested synchs
   double last_wall_time;
   time_sink working_on, was_working_on;
   double times_spent[Other+1];
@@ -1009,13 +1060,12 @@ class fields {
                                            complex<double> kphase[8], int &ncopies) const;
   // step.cpp
   void phase_material();
-  void step_h();
-  void step_h_source();
-  void step_d();
-  void step_d_source();
+  void step_b_source(int including_integrated = 0);
+  void update_h_from_b();
+  void step_db(field_type ft);
+  void step_d_source(int including_integrated = 0);
   void update_e_from_d();
   void update_from_e();
-  void step_boundaries(field_type);
   void calc_sources(double tim);
   int cluster_some_bands_cleverly(double *tf, double *td, complex<double> *ta,
                                   int num_freqs, int fields_considered, int maxbands,
