@@ -1,5 +1,5 @@
 // -*- C++ -*-
-/* Copyright (C) 2006 Massachusetts Institute of Technology
+/* Copyright (C) 2005-2008 Massachusetts Institute of Technology
 %
 %  This program is free software; you can redistribute it and/or modify
 %  it under the terms of the GNU General Public License as published by
@@ -25,15 +25,15 @@ using namespace std;
 
 namespace meep {
 
-const int NUM_FIELD_COMPONENTS = 15;
-const int NUM_FIELD_TYPES = 4;
+const int NUM_FIELD_COMPONENTS = 20;
+const int NUM_FIELD_TYPES = 5;
 
 enum component { Ex=0, Ey, Er, Ep, Ez, Hx, Hy, Hr, Hp, Hz,
-                 Dx, Dy, Dr, Dp, Dz, Dielectric };
+                 Dx, Dy, Dr, Dp, Dz, Bx, By, Br, Bp, Bz, Dielectric, Permeability };
 enum derived_component { Sx=100, Sy, Sr, Sp, Sz, EnergyDensity,
 			 D_EnergyDensity, H_EnergyDensity };
 enum ndim { D1=0, D2, D3, Dcyl };
-enum field_type { E_stuff=0, H_stuff=1, D_stuff=2, P_stuff=3 };
+enum field_type { E_stuff=0, H_stuff=1, D_stuff=2, B_stuff=3, P_stuff=4 };
 enum boundary_side { High=0, Low };
 enum direction { X=0,Y,Z,R,P, NO_DIRECTION };
 struct signed_direction {
@@ -65,16 +65,23 @@ inline direction stop_at_direction(ndim dim) {
   return (direction) (dim + 1 + 2 * (dim == D1));
 }
 
+component first_field_component(field_type ft);
+
 #define FOR_FIELD_TYPES(ft) for (field_type ft = E_stuff; \
                                  ft <= P_stuff; ft = (field_type) (ft+1))
 #define FOR_ELECTRIC_COMPONENTS(c) for (component c = Ex; \
                                         c < Hx; c = (component) (c+1))
 #define FOR_MAGNETIC_COMPONENTS(c) for (component c = Hz; \
                                         c > Ez; c = (component) (c-1))
+#define FOR_B_COMPONENTS(c) for (component c = Bz; \
+				 c > Dz; c = (component) (c-1))
+#define FOR_H_AND_B(h,b) for (component h=Hx, b=Bx; \
+			      h <= Hz; h = (component) (h+1), b = (component) (b+1))
 #define FOR_D_COMPONENTS(c) for (component c = Dz; \
                                  c > Hz; c = (component) (c-1))
 #define FOR_E_AND_D(e,d) for (component e = Ex, d = Dx; \
                               e <= Ez; e = (component) (e+1), d = (component) (d+1))
+#define FOR_FT_COMPONENTS(ft,c) for (component c = first_field_component(ft), loop_cstop = component(first_field_component(ft) + NUM_FIELD_TYPES); c < loop_cstop; c = component(c+1))
 #define FOR_COMPONENTS(c) for (component c = Ex,loop_stop_co=Ey; \
                                c != loop_stop_co; \
                                c = (component)((c+1)%NUM_FIELD_COMPONENTS), \
@@ -88,10 +95,14 @@ inline direction stop_at_direction(ndim dim) {
 			  s = (boundary_side) ((s+1) % 2), \
 			  loop_stop_bi = High)
 
+// only loop over directions where we have coordinates
 #define LOOP_OVER_DIRECTIONS(dim, d) for (direction d = start_at_direction(dim), \
                                      loop_stop_directi = stop_at_direction(dim); \
                                      d < loop_stop_directi; d = (direction) (d+1))
 
+// loop over all directions in which we might have fields
+#define LOOP_OVER_FIELD_DIRECTIONS(dim, d) for (direction d = dim == Dcyl ? Z : X; d < (dim == Dcyl ? NO_DIRECTION : R); d = direction(d+1))
+
 // loop over indices idx from is to ie (inclusive) in v
 #define LOOP_OVER_IVECS(v, is, ie, idx) \
   for (int loop_is1 = (is).yucky_val(0), \
@@ -120,6 +131,9 @@ inline direction stop_at_direction(ndim dim) {
 #define LOOP_OVER_VOL_OWNED(v, c, idx) \
   LOOP_OVER_IVECS(v, (v).little_owned_corner(c), (v).big_corner(), idx)
 
+#define LOOP_OVER_VOL_OWNED0(v, c, idx) \
+  LOOP_OVER_IVECS(v, (v).little_owned_corner0(c), (v).big_corner(), idx)
+
 #define LOOP_OVER_VOL_NOTOWNED(v, c, idx) \
  for (ivec loop_notowned_is((v).dim,0), loop_notowned_ie((v).dim,0); \
       loop_notowned_is == zero_ivec((v).dim);) \
@@ -162,6 +176,11 @@ inline bool has_direction(ndim dim, direction d) {
   return false;
 }
 
+inline bool has_field_direction(ndim dim, direction d) {
+  LOOP_OVER_FIELD_DIRECTIONS(dim, dd) if (dd == d) return true;
+  return false;
+}
+
 // true if d is polar while dim is cartesian, or vice versa 
 inline bool coordinate_mismatch(ndim dim, direction d) {
   return (d != NO_DIRECTION &&
@@ -173,7 +192,8 @@ extern void abort(const char *, ...); // mympi.cpp
 
 inline bool is_electric(component c) { return c < Hx; }
 inline bool is_magnetic(component c) { return c >= Hx && c < Dx; }
-inline bool is_D(component c) { return c >= Dx && c < Dielectric; }
+inline bool is_D(component c) { return c >= Dx && c < Bx; }
+inline bool is_B(component c) { return c >= Bx && c < Dielectric; }
 inline bool is_derived(int c) { return c >= Sx; }
 inline bool is_poynting(derived_component c) { return c < EnergyDensity; }
 inline bool is_energydensity(derived_component c) { return c>=EnergyDensity; }
@@ -181,6 +201,7 @@ inline field_type type(component c) {
   if (is_electric(c)) return E_stuff;
   else if (is_magnetic(c)) return H_stuff;
   else if (is_D(c)) return D_stuff;
+  else if (is_B(c)) return B_stuff;
   abort("Invalid field in type.\n");
   return E_stuff; // This is never reached.
 }
@@ -193,13 +214,14 @@ const char *dimension_name(ndim);
 direction component_direction(int c);
 int direction_component(int c, direction d);
 inline direction component_direction(component c) {
-  switch (c) {
-  case Ex: case Hx: case Dx: return X;
-  case Ey: case Hy: case Dy: return Y;
-  case Ez: case Hz: case Dz: return Z;
-  case Er: case Hr: case Dr: return R;
-  case Ep: case Hp: case Dp: return P;
+  switch (c) {    
+  case Ex: case Hx: case Dx: case Bx: return X;
+  case Ey: case Hy: case Dy: case By: return Y;
+  case Ez: case Hz: case Dz: case Bz: return Z;
+  case Er: case Hr: case Dr: case Br: return R;
+  case Ep: case Hp: case Dp: case Bp: return P;
   case Dielectric: return NO_DIRECTION;
+  case Permeability: return NO_DIRECTION;
   }
   return X; // This code is never reached...
 }
@@ -226,7 +248,9 @@ inline component direction_component(component c, direction d) {
   if (is_electric(c)) start_point = Ex;
   else if (is_magnetic(c)) start_point = Hx;
   else if (is_D(c)) start_point = Dx;
+  else if (is_B(c)) start_point = Bx;
   else if (c == Dielectric && d == NO_DIRECTION) return Dielectric;
+  else if (c == Permeability && d == NO_DIRECTION) return Permeability;
   else abort("unknown field component %d", c);
   switch (d) {
   case X: return start_point;
@@ -267,6 +291,17 @@ inline bool coordinate_mismatch(ndim dim, derived_component c) {
   return coordinate_mismatch(dim, component_direction(c));
 }
 
+// cyclically shift a direction d or a component c by shift
+// assumes: shift >= -99, {d, component_direction(c)} != NO_DIRECTION,
+//          and has_direction(dim, {d, component_direction(c)})
+inline direction cycle_direction(ndim dim, direction d, int shift) {
+  int start = dim == Dcyl ? 2 : 0;
+  return direction((d - start + shift + 99) % 3 + start);
+}
+inline component cycle_component(ndim dim, component c, int shift) {
+  return direction_component(c, cycle_direction(dim, component_direction(c), shift));
+}
+
 class vec;
 vec veccyl(double rr, double zz);
 vec zero_vec(ndim);
@@ -610,7 +645,7 @@ class volume {
   int nz() const { return num_direction(Z); }
 
   bool has_field(component c) const {
-    if (dim == D1) return c == Ex || c == Hy || c == Dx;
+    if (dim == D1) return c == Ex || c == Hy || c == Dx || c == By;
     return (dim == Dcyl)?component_direction(c)>Y:component_direction(c)<R;
   }
   int has_boundary(boundary_side,direction) const;
@@ -664,6 +699,13 @@ class volume {
 
   bool contains(const vec &) const;
   bool contains(const ivec &) const;
+
+  /* differs from little_owned_corner in that it doesn't count
+     "ownership" of the r=0 origin for Dcyl, which is updated separately */
+  ivec little_owned_corner0(component c) const {
+    return ivec(little_corner() + one_ivec(dim)*2 - iyee_shift(c));
+  }
+
   ivec little_owned_corner(component c) const;
   bool owns(const ivec &) const;
   geometric_volume surroundings() const;
@@ -693,9 +735,9 @@ class volume {
   ivec iyee_shift(component c) const {
     ivec out = zero_ivec(dim);
     LOOP_OVER_DIRECTIONS(dim,d)
-      if (c == Dielectric ||
+      if (c == Dielectric || c == Permeability || 
           ((is_electric(c) || is_D(c)) && d == component_direction(c)) ||
-          (is_magnetic(c) && d != component_direction(c)))
+          ((is_magnetic(c) || is_B(c)) && d != component_direction(c)))
         out.set_direction(d,1);
     return out;
   }
@@ -709,6 +751,7 @@ class volume {
   void set_origin(direction d, int o);
   void center_origin(void) { shift_origin(-icenter()); }
   double origin_in_direction(direction d) const{return origin.in_direction(d);}
+  int iorigin_in_direction(direction d) const{return io.in_direction(d);}
   double origin_r() const { return origin.r(); }
   double origin_x() const { return origin.x(); }
   double origin_y() const { return origin.y(); }
@@ -778,11 +821,6 @@ class symmetry {
   friend symmetry r_to_minus_r_symmetry(double m);
 };
 
-symmetry identity();
-symmetry rotate4(direction,const volume &);
-symmetry rotate2(direction,const volume &);
-symmetry mirror(direction,const volume &);
-
 class geometric_volume_list {
 public:
   geometric_volume_list(const geometric_volume &gv, int c, complex<double> weight = 1.0, geometric_volume_list *next = 0) : gv(gv), c(c), weight(weight), next(next) {}
