@@ -1,4 +1,4 @@
-/* Copyright (C) 2006 Massachusetts Institute of Technology
+/* Copyright (C) 2005-2008 Massachusetts Institute of Technology
 %
 %  This program is free software; you can redistribute it and/or modify
 %  it under the terms of the GNU General Public License as published by
@@ -119,6 +119,8 @@ complex<double> fields::get_field(component c, const vec &loc) const {
   switch (c) {
   case Dielectric:
     return get_eps(loc);
+  case Permeability:
+    abort("get_field(Permeability) unimplemented");
   default:
     ivec ilocs[8];
     double w[8];
@@ -158,7 +160,7 @@ complex<double> fields_chunk::get_field(component c, const ivec &iloc) const {
    It is *not* necessarily non-overlapping with other chunks. */
 geometric_volume fields_chunk::get_field_gv(component c) const {
   switch (c) {
-  case Dielectric:
+  case Dielectric: case Permeability:
     c = v.eps_component();
   default:
     return geometric_volume(v.loc(c, 0), v.loc(c, v.ntot() - 1));
@@ -171,6 +173,7 @@ complex<double> fields_chunk::get_field(component c, const vec &loc) const {
   ivec ilocs[8];
   double w[8];
   switch (c) {
+  case Permeability: abort("non-collective get_field(mu) unimplemented");
   case Dielectric: {
     c = v.eps_component();
     v.interpolate(c, loc, ilocs, w);
@@ -430,12 +433,8 @@ complex<double> monitor_point::get_component(component w) {
 }
   
 double monitor_point::poynting_in_direction(direction d) {
-  int direction_shift = 0;
-  if (loc.dim == Dcyl)
-    direction_shift = 2;
-
-  direction d1 = (direction) (((d+1) - direction_shift)%3 + direction_shift);
-  direction d2 = (direction) (((d+2) - direction_shift)%3 + direction_shift);
+  direction d1 = cycle_direction(loc.dim, d, 1);
+  direction d2 = cycle_direction(loc.dim, d, 2);
 
   // below Ex and Hx are used just to say that we want electric or magnetic component
   complex<double> E1 = get_component(direction_component(Ex, d1));
