@@ -1,4 +1,4 @@
-/* Copyright (C) 2006 Massachusetts Institute of Technology
+/* Copyright (C) 2005-2008 Massachusetts Institute of Technology
 %
 %  This program is free software; you can redistribute it and/or modify
 %  it under the terms of the GNU General Public License as published by
@@ -47,11 +47,11 @@ structure::structure(const volume &thev, material_function &eps,
   Courant(Courant), gv(D1) // Aaack, this is very hokey.
 {
   outdir = ".";
-  choose_chunkdivision(thev, num == 0 ? count_processors() : num, br, s);
+  choose_chunkdivision(thev, num, br, s);
   set_materials(eps, use_anisotropic_averaging, tol, maxeval);
 }
 
-structure::structure(const volume &thev, double eps(const vec &), 
+structure::structure(const volume &thev, double eps(const vec &),
 		     const boundary_region &br,
 		     const symmetry &s,
 		     int num, double Courant, bool use_anisotropic_averaging,
@@ -59,7 +59,7 @@ structure::structure(const volume &thev, double eps(const vec &),
   Courant(Courant), gv(D1) // Aaack, this is very hokey.
 {
   outdir = ".";
-  choose_chunkdivision(thev, num == 0 ? count_processors() : num, br, s);
+  choose_chunkdivision(thev, num, br, s);
   simple_material_function epsilon(eps);
   set_materials(epsilon, use_anisotropic_averaging, tol, maxeval);
 }
@@ -69,6 +69,8 @@ void structure::choose_chunkdivision(const volume &thev,
 				     const boundary_region &br,
 				     const symmetry &s) {
   user_volume = thev;
+  if (desired_num_chunks == 0)
+    desired_num_chunks = count_processors();
   if (thev.dim == Dcyl && thev.get_origin().r() < 0)
     abort("r < 0 origins are not supported");
   v = thev;
@@ -86,7 +88,7 @@ void structure::choose_chunkdivision(const volume &thev,
       for (int n=0;n<S.multiplicity();n++)
         if (has_direction(thev.dim,d) &&
             (S.transform(d,n).d != d || S.transform(d,n).flipped)) {
-          if (thev.num_direction(d) & 1 && !break_this[d])
+          if (thev.num_direction(d) & 1 && !break_this[d] && !quiet)
             master_printf("Padding %s to even number of grid points.\n",
 			  direction_name(d));
           break_this[dd] = true;
@@ -97,8 +99,9 @@ void structure::choose_chunkdivision(const volume &thev,
       if (break_mult == S.multiplicity()) break_this[d] = false;
       if (break_this[d]) {
 	break_mult *= 2;
-	master_printf("Halving computational cell along direction %s\n",
-		      direction_name(direction(d)));
+	if (!quiet)
+	  master_printf("Halving computational cell along direction %s\n",
+			direction_name(direction(d)));
 	v = v.halve((direction)d);
       }
     }
@@ -143,7 +146,7 @@ void boundary_region::apply(structure *s) const {
       && s->user_volume.num_direction(d) > 1) {
     switch (kind) {
     case NOTHING_SPECIAL: break;
-    case PML: s->use_pml(d, side, thickness, strength); break;
+    case PML: s->use_pml(d, side, thickness); break;
     default: abort("unknown boundary region kind");
     }
   }
@@ -153,12 +156,13 @@ void boundary_region::apply(structure *s) const {
 
 void boundary_region::apply(const structure *s, structure_chunk *sc) const {
   if (has_direction(s->v.dim, d) && s->user_volume.has_boundary(side, d)
-      && s->user_volume.num_direction(d) > 1) {
+      && s->user_volume.num_direction(d) > 1) {    
     switch (kind) {
     case NOTHING_SPECIAL: break;
     case PML: 
       sc->use_pml(d, thickness, s->user_volume.boundary_location(side, d),
-		  strength); 
+		  Rasymptotic,
+		  pml_profile, pml_profile_data, pml_profile_integral); 
       break;
     default: abort("unknown boundary region kind");
     }
@@ -167,8 +171,12 @@ void boundary_region::apply(const structure *s, structure_chunk *sc) const {
     next->apply(s, sc);
 }
 
+double pml_quadratic_profile(double u, void *d) { (void)d; return u * u; }
+
 boundary_region pml(double thickness, direction d, boundary_side side) {
-  return boundary_region(boundary_region::PML, thickness, 1.0, d, side, NULL);
+  return boundary_region(boundary_region::PML, thickness, 
+			 1e-15, pml_quadratic_profile, NULL, 1./3., 
+			 d, side, NULL);
 }
 boundary_region pml(double thickness, direction d) {
   return (pml(thickness, d, Low) + pml(thickness, d, High));
@@ -315,6 +323,9 @@ void structure::set_materials(material_function &mat,
 			      bool use_anisotropic_averaging,
 			      double tol, int maxeval) {
   set_epsilon(mat, use_anisotropic_averaging, tol, maxeval);
+  if (mat.has_mu()) set_mu(mat);
+  FOR_D_COMPONENTS(c) if (mat.has_conductivity(c)) set_conductivity(c, mat);
+  FOR_B_COMPONENTS(c) if (mat.has_conductivity(c)) set_conductivity(c, mat);
   if (mat.has_chi3()) set_chi3(mat);
   if (mat.has_chi2()) set_chi2(mat);
 }
@@ -338,6 +349,37 @@ void structure::set_epsilon(double eps(const vec &),
   set_epsilon(epsilon, use_anisotropic_averaging, tol, maxeval);
 }
 
+void structure::set_mu(material_function &mu) {
+  double tstart = wall_time();
+  changing_chunks();
+  for (int i=0;i<num_chunks;i++)
+    if (chunks[i]->is_mine())
+      chunks[i]->set_mu(mu);
+  if (!quiet)
+    master_printf("time for set_mu = %g s\n", wall_time() - tstart);
+}
+
+void structure::set_mu(double mufunc(const vec &)) {
+  simple_material_function mu(mufunc);
+  set_mu(mu);
+}
+
+void structure::set_conductivity(component c, material_function &C) {
+  if (!v.has_field(c)) return;
+  double tstart = wall_time();
+  changing_chunks();
+  for (int i=0;i<num_chunks;i++)
+    if (chunks[i]->is_mine())
+      chunks[i]->set_conductivity(c, C);
+  if (!quiet)
+    master_printf("time for set_conductivity = %g s\n", wall_time() - tstart);
+}
+
+void structure::set_conductivity(component c, double Cfunc(const vec &)) {
+  simple_material_function conductivity(Cfunc);
+  set_conductivity(c, conductivity);
+}
+
 void structure::set_chi3(material_function &eps) {
   changing_chunks();
   for (int i=0;i<num_chunks;i++)
@@ -362,9 +404,8 @@ void structure::set_chi2(double eps(const vec &)) {
   set_chi2(epsilon);
 }
 
-void structure::use_pml(direction d, boundary_side b, double dx,
-			double strength) {
-  if (strength == 0.0 || dx <= 0.0) return;
+void structure::use_pml(direction d, boundary_side b, double dx) {
+  if (dx <= 0.0) return;
   volume pml_volume = v;
   pml_volume.set_num_direction(d, int(dx*user_volume.a + 1 + 0.5)); //FIXME: exact value?
   if (b == High)
@@ -387,28 +428,77 @@ void structure::mix_with(const structure *oth, double f) {
 }
 
 structure_chunk::~structure_chunk() {
-  FOR_ELECTRIC_COMPONENTS(c) {
-    FOR_DIRECTIONS(d)
+  FOR_COMPONENTS(c) {
+    FOR_DIRECTIONS(d) {
       delete[] inveps[c][d];
-    delete[] chi3[c];
+      delete[] invmu[c][d];
+      delete[] conductivity[c][d];
+      delete[] condinv[c][d];
+    }
     delete[] chi2[c];
+    delete[] chi3[c];
   }
   delete[] eps;
-
-  FOR_COMPONENTS(c) FOR_DIRECTIONS(d) delete[] C[d][c];
-  FOR_COMPONENTS(c)
-    FOR_DIRECTIONS(d) FOR_DIRECTIONS(d2)
-        delete[] Cdecay[d][c][d2];
+  FOR_DIRECTIONS(d) { 
+    delete[] sig[d];
+    delete[] siginv[d];
+  }
   if (pb) delete pb;
 }
 
 void structure_chunk::mix_with(const structure_chunk *n, double f) {
   for (int i=0;i<v.ntot();i++)
     eps[i] = 1.0/(1.0/eps[i] + f*(1.0/n->eps[i]-1.0/eps[i]));
-  FOR_ELECTRIC_COMPONENTS(c) FOR_DIRECTIONS(d)
-    if (inveps[c][d])
-      for (int i=0;i<v.ntot();i++)
-        inveps[c][d][i] += f*(n->inveps[c][d][i] - inveps[c][d][i]);
+  FOR_COMPONENTS(c) FOR_DIRECTIONS(d) {
+    if (!inveps[c][d] && n->inveps[c][d]) {
+      inveps[c][d] = new double[v.ntot()];
+      if (component_direction(c) == d) // diagonal components = 1 by default
+	for (int i=0;i<v.ntot();i++) inveps[c][d][i] = 1.0;
+      else
+	for (int i=0;i<v.ntot();i++) inveps[c][d][i] = 0.0;
+    }
+    if (!invmu[c][d] && n->invmu[c][d]) {
+      invmu[c][d] = new double[v.ntot()];
+      if (component_direction(c) == d) // diagonal components = 1 by default
+	for (int i=0;i<v.ntot();i++) invmu[c][d][i] = 1.0;
+      else
+	for (int i=0;i<v.ntot();i++) invmu[c][d][i] = 0.0;
+    }
+    if (!conductivity[c][d] && n->conductivity[c][d]) {
+      conductivity[c][d] = new double[v.ntot()];
+      for (int i=0;i<v.ntot();i++) conductivity[c][d][i] = 0.0;
+    }
+    if (inveps[c][d]) {
+      if (n->inveps[c][d])
+	for (int i=0;i<v.ntot();i++)
+	  inveps[c][d][i] += f*(n->inveps[c][d][i] - inveps[c][d][i]);
+      else {
+	double nval = component_direction(c) == d ? 1.0 : 0.0; // default
+	for (int i=0;i<v.ntot();i++)
+	  inveps[c][d][i] += f*(nval - inveps[c][d][i]);
+      }
+    }
+    if (invmu[c][d]) {
+      if (n->invmu[c][d])
+	for (int i=0;i<v.ntot();i++)
+	  invmu[c][d][i] += f*(n->invmu[c][d][i] - invmu[c][d][i]);
+      else {
+	double nval = component_direction(c) == d ? 1.0 : 0.0; // default
+	for (int i=0;i<v.ntot();i++)
+	  invmu[c][d][i] += f*(nval - invmu[c][d][i]);
+      }
+    }
+    if (conductivity[c][d]) {
+      if (n->conductivity[c][d])
+	for (int i=0;i<v.ntot();i++)
+	  conductivity[c][d][i] += f*(n->conductivity[c][d][i]
+				      - conductivity[c][d][i]);
+      else
+	for (int i=0;i<v.ntot();i++)
+	  conductivity[c][d][i] += f*(0.0 - conductivity[c][d][i]);
+    }
+    condinv_stale = true;
+  }
   // Mix in the polarizability...
   polarizability *po = pb, *pn = n->pb;
   while (po && pn) {
@@ -421,77 +511,78 @@ void structure_chunk::mix_with(const structure_chunk *n, double f) {
   }
 }
 
-const double Cmax = 0.5;
-
 void structure_chunk::use_pml(direction d, double dx, double bloc,
-			      double strength) {
-  if (strength == 0.0 || dx <= 0.0) return;
-  const double prefac = strength * Cmax/(dx*dx);
+			      double Rasymptotic,
+			      pml_profile_func pml_profile,
+			      void *pml_profile_data,
+			      double pml_profile_integral) {
+  if (dx <= 0.0) return;
+  const double prefac = (-log(Rasymptotic))/(4*dx*pml_profile_integral);
   // Don't bother with PML if we don't even overlap with the PML region...
-  if (bloc > v.boundary_location(High,d) + dx + 1.0/a ||
-      bloc < v.boundary_location(Low,d) - dx - 1.0/a) return;
+  if (bloc > v.boundary_location(High,d) + dx + 1.0/a - 1e-10 ||
+      bloc < v.boundary_location(Low,d) - dx - 1.0/a + 1e-10) return;
   if (is_mine()) {
-    FOR_COMPONENTS(c)
-      if (v.has_field(c) && component_direction(c) != d) {
-        if (!C[d][c]) {
-          C[d][c] = new double[v.ntot()];
-          for (int i=0;i<v.ntot();i++) C[d][c][i] = 0.0;
-        }
-	LOOP_OVER_VOL(v, (component) c, i) {
-	  IVEC_LOOP_LOC(v, here);
-          const double x =
-            0.5/a*((int)(dx*(2*a)+0.5) -
-                   (int)(2*a*fabs(bloc-here.in_direction(d))+0.5));
-          if (x > 0) C[d][c][i] = prefac*x*x;
-        }
-      }
-  }
-}
-
-void structure_chunk::update_pml_arrays() {
-  if (!is_mine()) return;
-  FOR_DIRECTIONS(d)
-    FOR_COMPONENTS(c) 
-      if (C[d][c] != NULL) {
-	bool all_zeros = true;
-	LOOP_OVER_VOL_OWNED(v, c, i) {
-	  if (C[d][c][i] != 0.0) {
-	    all_zeros = false;
-	    goto done; // can't use 'break' since LOOP has nested loops
-	  }
-	}
-      done:
-	if (all_zeros) {
-	  delete[] C[d][c];
-	  C[d][c] = NULL;
+    if (sig[d]) {
+      delete[] sig[d]; 
+      delete[] siginv[d];
+      sig[d] = NULL;
+      siginv[d] = NULL;
+    }
+    LOOP_OVER_FIELD_DIRECTIONS(v.dim, dd) {
+      if (!sig[dd]) {
+	int spml = (dd==d)?(2*v.num_direction(d)+2):1;
+	sigsize[dd] = spml;
+	sig[dd] = new double[spml];
+	siginv[dd] = new double[spml];
+	for (int i=0;i<spml;++i) {
+	  sig[dd][i] = 0.0;
+	  siginv[dd][i] = 1.0;
 	}
       }
-  update_Cdecay();
+    }
+    for (int i=v.little_corner().in_direction(d);
+	 i<=v.big_corner().in_direction(d)+1;++i) {
+      int idx = i - v.little_corner().in_direction(d);
+      double here = i * 0.5/a;
+      const double x =
+	0.5/a*((int)(dx*(2*a)+0.5) - (int)(fabs(bloc-here)*(2*a)+0.5));
+      if (x > 0) {
+	sig[d][idx]=0.5*dt*prefac*pml_profile(x/dx, pml_profile_data);
+	siginv[d][idx] = 1/(1+sig[d][idx]);	
+      }
+    }
+  }
+  condinv_stale = true;
 }
 
-void structure_chunk::update_Cdecay() {
-  FOR_DIRECTIONS(d) FOR_COMPONENTS(c) FOR_DIRECTIONS(d2) {
-    delete[] Cdecay[d][c][d2];
-    Cdecay[d][c][d2] = NULL;
-  }
-  FOR_DIRECTIONS(d) FOR_COMPONENTS(c) 
-    if (C[d][c] != NULL)
-      FOR_DIRECTIONS(d2) 
-	if ((inveps[c][d2] || d2 == component_direction(c)) && d2 != d) {
-	  Cdecay[d][c][d2] = new double[v.ntot()];
-	  for (int i=0;i<v.ntot();i++) {
-	    if (is_magnetic(c)) Cdecay[d][c][d2][i] = 1.0/(1.0+0.5*C[d][c][i]);
-	    else if (is_electric(c))
-          Cdecay[d][c][d2][i] = inveps[c][d2][i]/(1.0+0.5*C[d][c][i]*inveps[c][d2][i]);
-        else Cdecay[d][c][d2][i] = 1.0/(1.0+0.5*C[d][c][i]); // FIXME: Is this right?
-	    /*
-	      TODO: THIS IS NOT CORRECT FOR NON-DIAGONAL INVEPS...maybe
-              above code is not correct either????
-	    if (Cdecay[d][c][d2][i] == 0.0)
-	      abort("In update_Cdecay: Cdecay == 0\n");
-	    */
-	  }
+void structure_chunk::update_condinv() {
+  if (!condinv_stale || !is_mine()) return;
+  FOR_COMPONENTS(c) {
+    direction d = component_direction(c);
+    if (conductivity[c][d]) {
+      if (!condinv[c][d]) condinv[c][d] = new double[v.ntot()];
+      const direction dsig = cycle_direction(v.dim,d,1);
+      const bool have_pml = sigsize[dsig] > 1;
+      if (!have_pml) {
+	LOOP_OVER_VOL(v, c, i)
+	  condinv[c][d][i] = 1 / (1 + conductivity[c][d][i] * dt * 0.5);
+      }
+      else { // include PML conductivity in condinv
+	int k0 = v.little_corner().in_direction(dsig);
+	LOOP_OVER_VOL(v, c, i) {
+	  IVEC_LOOP_ILOC(v, iloc);
+          int k = iloc.in_direction(dsig) - k0;
+	  condinv[c][d][i] = 1 / (1 + conductivity[c][d][i] * dt * 0.5
+				  + sig[dsig][k]);
 	}
+      }
+    }
+    else if (condinv[c][d]) { // condinv not needed
+      delete[] condinv[c][d];
+      condinv[c][d] = NULL;
+    }
+  }
+  condinv_stale = false;
 }
 
 structure_chunk::structure_chunk(const structure_chunk *o) : gv(o->gv) {
@@ -527,29 +618,42 @@ structure_chunk::structure_chunk(const structure_chunk *o) : gv(o->gv) {
       chi2[c] = NULL;
     }
   }
-  FOR_COMPONENTS(c) FOR_DIRECTIONS(d)
-    if (is_mine() && o->inveps[c][d]) {
+  FOR_COMPONENTS(c) FOR_DIRECTIONS(d) if (is_mine()) {
+    if (o->inveps[c][d]) {
       inveps[c][d] = new double[v.ntot()];
-      for (int i=0;i<v.ntot();i++) inveps[c][d][i] = o->inveps[c][d][i];
-    } else {
-      inveps[c][d] = NULL;
-    }
-  // Allocate the conductivity arrays:
-  FOR_DIRECTIONS(d) FOR_COMPONENTS(c) C[d][c] = NULL;
-  FOR_DIRECTIONS(d) FOR_COMPONENTS(c) FOR_DIRECTIONS(d2)
-    Cdecay[d][c][d2] = NULL;
-  // Copy over the conductivity arrays:
+      memcpy(inveps[c][d], o->inveps[c][d], v.ntot()*sizeof(double));
+    } else inveps[c][d] = NULL;
+    if (o->invmu[c][d]) {
+      invmu[c][d] = new double[v.ntot()];
+      memcpy(invmu[c][d], o->invmu[c][d], v.ntot()*sizeof(double));
+    } else invmu[c][d] = NULL;
+    if (o->conductivity[c][d]) {
+      conductivity[c][d] = new double[v.ntot()];
+      memcpy(conductivity[c][d], o->conductivity[c][d],
+	     v.ntot()*sizeof(double));
+      condinv[c][d] = new double[v.ntot()];
+      memcpy(condinv[c][d], o->condinv[c][d], v.ntot()*sizeof(double));
+    } else conductivity[c][d] = condinv[c][d] = NULL;
+  }
+  condinv_stale = o->condinv_stale;
+  // Allocate the PML conductivity arrays:
+  FOR_DIRECTIONS(d) { 
+    sig[d] = NULL; 
+    siginv[d] = NULL;
+    sigsize[d] = 0;
+  }
+  for (int i=0;i<5;++i) sigsize[i] = 0;
+  // Copy over the PML conductivity arrays:
   if (is_mine())
-    FOR_DIRECTIONS(d) FOR_COMPONENTS(c)
-      if (o->C[d][c]) {
-        C[d][c] = new double[v.ntot()];
-        for (int i=0;i<v.ntot();i++) C[d][c][i] = o->C[d][c][i];
-        FOR_DIRECTIONS(d2)
-          if (o->Cdecay[d][c][d2]) {
-            Cdecay[d][c][d2] = new double[v.ntot()];
-            for (int i=0;i<v.ntot();i++)
-              Cdecay[d][c][d2][i] = o->Cdecay[d][c][d2][i];
-          }
+    FOR_DIRECTIONS(d) 
+      if (o->sig[d]) {
+	sig[d] = new double[2*v.num_direction(d)+1];
+	siginv[d] = new double[2*v.num_direction(d)+1];
+	sigsize[d] = o->sigsize[d];
+	for (int i=0;i<2*v.num_direction(d)+1;i++) {
+	  sig[d][i] = o->sig[d][i];
+	  siginv[d][i] = o->sig[d][i];
+	}
       }
 }
 
@@ -561,16 +665,24 @@ void structure_chunk::set_chi3(material_function &epsilon) {
   FOR_ELECTRIC_COMPONENTS(c)
     if (inveps[c][component_direction(c)]) {
       if (!chi3[c]) chi3[c] = new double[v.ntot()];
+      bool trivial = true;
       LOOP_OVER_VOL(v, c, i) {
 	IVEC_LOOP_LOC(v, here);
         chi3[c][i] = epsilon.chi3(here);
+	trivial = trivial && (chi3[c][i] == 0.0);
       }
       
       /* currently, our update_e_from_d routine requires that
 	 chi2 be present if chi3 is, and vice versa */
       if (!chi2[c]) {
-	chi2[c] = new double[v.ntot()]; 
-	memset(chi2[c], 0, v.ntot() * sizeof(double)); // chi2 = 0 by default
+	if (!trivial) {
+	  chi2[c] = new double[v.ntot()]; 
+	  memset(chi2[c], 0, v.ntot() * sizeof(double)); // chi2 = 0
+	}
+	else { // no chi3, and chi2 is trivial (== 0), so delete
+	  delete[] chi3[c];
+	  chi3[c] = NULL;
+	}
       }
     }
 }
@@ -583,20 +695,67 @@ void structure_chunk::set_chi2(material_function &epsilon) {
   FOR_ELECTRIC_COMPONENTS(c)
     if (inveps[c][component_direction(c)]) {
       if (!chi2[c]) chi2[c] = new double[v.ntot()];
+      bool trivial = true;
       LOOP_OVER_VOL(v, c, i) {
 	IVEC_LOOP_LOC(v, here);
         chi2[c][i] = epsilon.chi2(here);
+	trivial = trivial && (chi2[c][i] == 0.0);
       }
 
       /* currently, our update_e_from_d routine requires that
 	 chi3 be present if chi2 is, and vice versa */
       if (!chi3[c]) {
-	chi3[c] = new double[v.ntot()]; 
-	memset(chi3[c], 0, v.ntot() * sizeof(double)); // chi3 = 0 by default
+	if (!trivial) {
+	  chi3[c] = new double[v.ntot()]; 
+	  memset(chi3[c], 0, v.ntot() * sizeof(double)); // chi3 = 0
+	}
+	else { // no chi2, and chi3 is trivial (== 0), so delete 
+	  delete[] chi2[c];
+          chi2[c] = NULL;
+	}
       }
     }
 }
 
+void structure_chunk::set_conductivity(component c, material_function &C) {
+  if (!is_mine() || !v.has_field(c)) return;
+
+  C.set_volume(v.pad().surroundings());
+
+  if (!is_electric(c) && !is_magnetic(c) && !is_D(c) && !is_B(c))
+    abort("invalid component for conductivity");
+
+  direction c_d = component_direction(c);
+  component c_C = is_electric(c) ? direction_component(Dx, c_d) :
+    (is_magnetic(c) ? direction_component(Bx, c_d) : c);
+  double *multby = is_electric(c) ? inveps[c][c_d] :
+    (is_magnetic(c) ? invmu[c][c_d] : NULL);
+  if (!conductivity[c_C][c_d]) 
+    conductivity[c_C][c_d] = new double[v.ntot()]; 
+  if (!conductivity[c_C][c_d]) abort("Memory allocation error.\n");
+  bool trivial = true;
+  double *cnd = conductivity[c_C][c_d];
+  if (multby) {
+    LOOP_OVER_VOL(v, c_C, i) {
+      IVEC_LOOP_LOC(v, here);
+      cnd[i] = C.conductivity(c, here) * multby[i];
+      trivial = trivial && (cnd[i] == 0.0);
+    }
+  }
+  else {
+    LOOP_OVER_VOL(v, c_C, i) {
+      IVEC_LOOP_LOC(v, here);
+      cnd[i] = C.conductivity(c, here);
+      trivial = trivial && (cnd[i] == 0.0);
+    }
+  }
+  if (trivial) { // skip conductivity computations if conductivity == 0
+    delete[] conductivity[c_C][c_d];
+    conductivity[c_C][c_d] = NULL;
+  }
+  condinv_stale = true;
+}
+
 structure_chunk::structure_chunk(const volume &thev, 
 				 const geometric_volume &vol_limit, 
 				 double Courant, int pr)
@@ -614,9 +773,18 @@ structure_chunk::structure_chunk(const volume &thev,
   eps = NULL;
   FOR_COMPONENTS(c) chi3[c] = NULL;
   FOR_COMPONENTS(c) chi2[c] = NULL;
-  FOR_COMPONENTS(c) FOR_DIRECTIONS(d) inveps[c][d] = NULL;
-  FOR_DIRECTIONS(d) FOR_COMPONENTS(c) C[d][c] = NULL;
-  FOR_DIRECTIONS(d) FOR_DIRECTIONS(d2) FOR_COMPONENTS(c) Cdecay[d][c][d2] = NULL;
+  FOR_COMPONENTS(c) FOR_DIRECTIONS(d) {
+    inveps[c][d] = NULL;
+    invmu[c][d] = NULL;
+    conductivity[c][d] = NULL;
+    condinv[c][d] = NULL;
+  }
+  condinv_stale = false;
+  FOR_DIRECTIONS(d) { 
+    sig[d] = NULL; 
+    siginv[d] = NULL;
+    sigsize[d] = 0;
+  }
 }
 
 double structure::max_eps() const {
