@@ -1,4 +1,4 @@
-/* Copyright (C) 2006 Massachusetts Institute of Technology
+/* Copyright (C) 2005-2008 Massachusetts Institute of Technology
 %
 %  This program is free software; you can redistribute it and/or modify
 %  it under the terms of the GNU General Public License as published by
@@ -85,7 +85,13 @@ const char *component_name(component c) {
   case Dz: return "dz";
   case Dr: return "dr";
   case Dp: return "dp";
+  case Bx: return "bx";
+  case By: return "by";
+  case Bz: return "bz";
+  case Br: return "br";
+  case Bp: return "bp";
   case Dielectric: return "eps";
+  case Permeability: return "mu";
   }
   return "Error in component_name";
 }
@@ -110,6 +116,16 @@ const char *component_name(int c) {
 	  : component_name(component(c)));
 }
 
+component first_field_component(field_type ft) {
+  switch (ft) {
+  case E_stuff: return Ex;
+  case H_stuff: return Hx;
+  case D_stuff: return Dx;
+  case B_stuff: return Bx;
+  default: abort("bug - only E/H/D/B stuff have components"); 
+  }
+}
+
 vec min(const vec &vec1, const vec &vec2) {
   vec m(vec1.dim);
   LOOP_OVER_DIRECTIONS(vec1.dim, d)
@@ -402,7 +418,7 @@ bool volume::get_boundary_icorners(component c, int ib,
 
 // first "owned" point for c in volume (see also volume::owns)
 ivec volume::little_owned_corner(component c) const {
-  ivec iloc(little_corner() + one_ivec(dim)*2 - iyee_shift(c));
+  ivec iloc(little_owned_corner0(c));
   if (dim == Dcyl && origin.r() == 0.0 && iloc.r() == 2)
     iloc.set_direction(R, 0);
   return iloc;
@@ -1254,12 +1270,12 @@ int symmetry::transform(int c, int n) const {
 }
 
 complex<double> symmetry::phase_shift(component c, int n) const {
-  if (c == Dielectric) return 1.0;
+  if (c == Dielectric || c == Permeability) return 1.0;
   complex<double> phase = transform(component_direction(c),n).phase;
   // flip tells us if we need to flip the sign.  For vectors (E), it is
   // just this simple:
   bool flip = transform(component_direction(c),n).flipped;
-  if (is_magnetic(c)) {
+  if (is_magnetic(c) || is_B(c)) {
     // Because H is a pseudovector, here we have to figure out if the
     // transformation changes the handedness of the basis.
     bool have_one = false, have_two = false;
@@ -1403,7 +1419,7 @@ static double energy_fun(const complex<double> *fields,
 }
 
 field_rfunction derived_component_func(derived_component c, const volume &v,
-				       int &nfields, component cs[6]) {
+				       int &nfields, component cs[12]) {
   switch (c) {
   case Sx: case Sy: case Sz: case Sr: case Sp:
     switch (c) {
@@ -1429,9 +1445,9 @@ field_rfunction derived_component_func(derived_component c, const volume &v,
     if (c != D_EnergyDensity)
       FOR_MAGNETIC_COMPONENTS(c0) if (v.has_field(c0)) {
 	cs[nfields++] = c0;
-	cs[nfields++] = c0;
+	cs[nfields++] = direction_component(Bx, component_direction(c0));
       }
-    if (nfields > 6) abort("too many field components");
+    if (nfields > 12) abort("too many field components");
     return energy_fun;
 
   default: 
